// SPDX-License-Identifier: MPL-2.0
// (c) 2019 Pascal Monasse <pascal.monasse@enpc.fr>
// Make a palette image into a color image.
// If the palette file does not exist, create one and optionally output it.
// TODO: read and apply the PLTE chunk of PNG if present.

#include "io_png.h"
#include <map>
#include <fstream>
#include <sstream>
#include <iostream>
#include <string>
#include <cstdlib>

struct color { unsigned char r,g,b; };
const color WHITE = {255,255,255};

bool read_palette(const char* fileName, std::map<unsigned char,color>& palette){
    std::ifstream file(fileName);
    if(! file.is_open())
        return false;
    while(true) {
        std::string str;
        std::getline(file, str);
        if(! file.good()) break;
        std::istringstream s(str);
        int idx, r, g, b;
        s >> idx >> r >> g >> b;
        if(s.fail()) break;
        if(r<0 || r>255 || g<0 || g>255 || b<0 || b>255) {
            std::cerr << "Wrong palette: some RGB component is out of bounds"
                      << std::endl;
            return false;
        }
        color c = {(unsigned char)r,(unsigned char)g,(unsigned char)b};
        palette[(unsigned char)idx] = c;
    }
    file.close();
    std::cout << "Palette file " << fileName << ": read " << palette.size()
              << " colors" << std::endl;
    return true;
}

bool write_palette(const char* fileName,const std::map<unsigned char,color>& m){
    std::ofstream file(fileName);
    if(! file.is_open())
        return false;
    for(std::map<unsigned char,color>::const_iterator it=m.begin();
        it!= m.end(); ++it) {
        file << (int)it->first << ' '
             << (int)it->second.r << ' '
             << (int)it->second.g << ' '
             << (int)it->second.b << std::endl; 
    }
    if(file.fail())
        return false;
    std::cout << "Writing new palette file " << fileName
              << " with " << m.size() << " colors" << std::endl;
    return true;
}

int main(int argc, char** argv) {
    if(argc!=3 && argc!=4) {
        std::cerr << "Usage: " << argv[0]
                  << " codedMask.png mask.png [paletteInOut.txt]" << std::endl;
        return 1;
    }
    size_t x, y;
    unsigned char* in = io_png_read_u8_gray(argv[1], &x, &y);
    if(!in) {
        std::cerr << "Error reading image file " << argv[1] << std::endl;
        return 1;
    }

    unsigned char* out = new unsigned char[3*x*y];

    std::map<unsigned char,color> palette;
    bool newPalette = (argc>3)? !read_palette(argv[3], palette): true;
    if(newPalette) {
        palette.clear(); // Ignore partially read file
        palette[0]=WHITE;
    }
    unsigned char nbCol=0;
    unsigned char *p=in, *q=out;
    for(size_t i=0; i<x*y; i++, p++, q++) {
        if(palette.find(*p) == palette.end()) {
            if(!newPalette) {
                std::cerr << "Error: index " << (int)*p
                          << " not found in palette" << std::endl;
                return 1;
            }
            int c = rand();
            color col = {(unsigned char)(c%255),
                         (unsigned char)((c>>8)%255),
                         (unsigned char)((c>>16)%255)};
            palette[*p] = col;
        }
        color c = palette[*p];
        q[0] = c.r; q[x*y]=c.g; q[2*x*y] = c.b;
    }
    if(io_png_write_u8(argv[2], out, x, y, 3) !=0) {
        std::cerr << "Error writing image file " << argv[2] << std::endl;
        return 1;
    }

    if(newPalette && argc>3 && !write_palette(argv[3],palette)) {
        std::cerr << "Error writing palette file " << argv[3] << std::endl;
        return 1;
    }

    free(in);
    delete [] out;
    return 0;
}
