// SPDX-License-Identifier: MPL-2.0
// (c) 2019-2020 Pascal Monasse <pascal.monasse@enpc.fr>
// Make a color image into a palette image: colors are assigned 8-bit index.
// Colors black and white are both associated to index 0, meaning no mask.
// TODO: write the PLTE chunk of PNG instead of a separate file.

#include "io_png.h"
#include <map>
#include <fstream>
#include <iostream>
#include <cstdlib>

struct color { unsigned char r,g,b; };
const color BLACK = {0,0,0};
const color WHITE = {255,255,255};

// Average colors
struct accumCol {
    int r,g,b; int n;
    accumCol(): r(0), g(0), b(0), n(0) {}
    void add(int r0, int g0, int b0) {
        ++n;
        r+=r0; g+=g0; b+=b0;
    }
    void normalize() {
        if(n==0) return;
        r/=n; g/=n; b/=n;
        n=1;
    }
};

// Necessary for std::map with color as key.
bool operator<(const color& c1, const color& c2) {
    return c1.r<c2.r|| (c1.r==c2.r && (c1.g<c2.g|| (c1.g==c2.g && c1.b<c2.b)));
}
bool operator!=(const color& c1, const color& c2) {
    return c1.r!=c2.r || c1.g!=c2.g || c1.b!=c2.b;
}

void labelcc(const unsigned char* in, size_t w, size_t h, int* cc, int x, int y,
             int label) {
    cc[y*w+x] = label;
    for(int dy=-1; dy<=1; dy++)
        for(int dx=-1; dx<=1; dx++) {
            int vx=x+dx, vy=y+dy, j=vy*w+vx;
            if(0<=vx && (size_t)vx<w && 0<=vy && (size_t)vy<h && cc[j]==0 &&
               (in[j]!=BLACK.r || in[j+w*h]!=BLACK.g || in[j+2*w*h]!=BLACK.b))
                labelcc(in, w, h, cc, vx, vy, label);
        }
}

int labelcc(const unsigned char* in, int* cc, size_t w, size_t h) {
    int n=1;
    for(size_t i=0; i<w*h; i++)
        cc[i]=0;
    const unsigned char* p=in;
    for(size_t y=0,i=0; y<h; y++)
        for(size_t x=0; x<w; x++, p++, i++) {
            color c = {*p, *(p+w*h), *(p+2*w*h)};
            if(c!=BLACK && cc[i]==0)
                labelcc(in, w, h, cc, (int)x, (int)y, n++);
        }
    return n;
}

int norm(color c) {
    return c.r*c.r+c.g*c.g+c.b*c.b;
}

int norm(unsigned char r, unsigned char g, unsigned char b) {
    color c={r,g,b};
    return norm(c);
}
 
// Mask images produced by the demo system are anti-aliased, creating plenty
// of small regions with intermediate color. We extract cc of the input and
// normalize to the color most distinct from black.
void realias(unsigned char* in, size_t w, size_t h) {
    int* cc = new int[w*h];
    int ncc = labelcc(in, cc, w, h);
    accumCol* col = new accumCol[ncc];
    //    for(int i=0; i<ncc; i++) col[i]=BLACK;
    const size_t n = w*h;
    for(size_t i=0; i<n; i++)
        col[cc[i]].add(in[i],in[i+n],in[i+2*n]);
        // if(norm(in[i],in[i+n],in[i+2*n]) > norm(col[cc[i]]))
        //     col[cc[i]] = {in[i],in[i+n],in[i+2*n]};
    for(int i=0; i<ncc; i++) col[i].normalize();
    for(size_t i=0; i<n; i++)
        in[i] = col[cc[i]].r;
    for(size_t i=0; i<n; i++)
        in[i+n] = col[cc[i]].g;
    for(size_t i=0; i<n; i++)
        in[i+2*n] = col[cc[i]].b;
    delete [] col;
    delete [] cc;
}

bool write_palette(const char* fileName,const std::map<color,unsigned char>& m){
    std::ofstream file(fileName);
    if(! file.is_open())
        return false;
    for(std::map<color,unsigned char>::const_iterator it=m.begin();
        it!= m.end(); ++it) {
        file << (int)it->second << ' '
             << (int)it->first.r << ' '
             << (int)it->first.g << ' '
             << (int)it->first.b << std::endl;
    }
    return !file.fail();
}

int main(int argc, char** argv) {
    if(argc!=3 && argc!=4) {
        std::cerr << "Usage: " << argv[0]
                  << " mask.png codedMask.png [paletteOut.txt]" << std::endl;
        return 1;
    }
    size_t x, y;
    unsigned char* in = io_png_read_u8_rgb(argv[1], &x, &y);
    if(!in) {
        std::cerr << "Error reading image file " << argv[1] << std::endl;
        return 1;
    }
    realias(in, x, y);

    const size_t n = x*y;
    unsigned char* out = new unsigned char[n];

    std::map<color,unsigned char> palette;
    palette[BLACK]=0;
    unsigned char nbCol=0;
    unsigned char *p=in, *q=out;
    for(size_t i=0; i<n; i++, p++, q++) {
        color c = {*p, *(p+n), *(p+2*n)};
        std::map<color,unsigned char>::iterator it=palette.find(c);
        if(it==palette.end()) {
            if(++nbCol == 0) {
                std::cerr << "More than 255 colors unsupported" << std::endl;
                return 1;
            }
            it = palette.insert(std::make_pair(c,nbCol)).first;
        }
        *q = it->second;
    }
    if(palette.size()<=1) {
        std::cout << "Uniform mask: no output file is written" << std::endl;
        return 0;
    }
    if(io_png_write_u8(argv[2], out, x, y, 1) !=0) {
        std::cerr << "Error writing image file " << argv[2] << std::endl;
        return 1;
    }

    if(argc>3 && !write_palette(argv[3], palette)) {
        std::cerr << "Error writing palette file " << argv[3] << std::endl;
        return 1;
    }

    free(in);
    delete [] out;
    return 0;
}
